//Driver for the rhombic dodecahedron lamp
#include <Adafruit_NeoPixel.h>

#define edge_count 24 //the number of edges on the shape
#define edge_size 7 //the amount of pixels per edge
#define led_count 168 //the total amount of leds in the shape
#define led_pin 6

#define animation_off 0 //all off
#define animation_basic_rainbow 1 //simple 7 pixel rainbow
#define animation_twinkles 2 //twinkels in defined palette
#define animation_traces 3 //traces across random edges
#define animation_droplets 4 //droplets in the middle of random edges
#define animation_cycler 5 //cycles from pixel 0 to max pixel with the next color of the palette

//basic rainbow projects a different segment of a rainbow on each edge
//twinkles heat up and cool down random pixels
//traces project random traces upon each edge

uint16_t animation_state = 0; //the state machine handling what pattern to show
uint8_t heat_animation = 0; //whether animations happen on heat or pixel color
uint8_t universal_cooldown = 1; //whether all pixels cool down over time
uint8_t universal_cooling = 5; //how much each pixel cools down each cycle

uint32_t animation_update_delay = 10000; //the delay on how often the animation is changed
uint32_t animation_update_target; //the target on when the animation updates

uint32_t pattern_update_delay = 100; //the delay on how often the pattern itself updates
uint8_t pattern_speed = 5;
uint32_t pattern_update_target; //the target on when the pattern updates

uint32_t frame_update_delay = 20; //the delay between frames. Determines frames per second and animation update delay
uint32_t frame_update_target; //the target on when the frame updates

uint32_t pixel_color[led_count]; //variable that stores the color of each pixel in 3x8 bits
uint8_t pixel_heat[led_count]; //variable that stores the pixel heat (0-255)

//The palette is a simplified colorscheme of 8 bits used to address each pixel. Animations can use 24 bit values calculated based on these 8 bit values
#define num_palettes 6 //the current amount of palettes
const uint8_t palette[num_palettes][8][3] = { //up to 8 colors per palette, in 24 bit color
  {{255, 0, 0}, {0, 255, 0}, {0, 0, 255}, {255, 255, 0}, {255, 0, 255}, {0, 255, 255}, {255, 255, 255}, {255, 80, 0}}, //palette 0, test all colors
  {{0, 240, 36}, {0, 240, 36}, {0, 160, 24}, {0, 160, 24}, {0, 88, 12}, {0, 88, 12}, {0, 88, 12}, {176, 4, 2}}, //palette 2 Holly
  {{48, 64, 72}, {48, 64, 72}, {48, 64, 72}, {48, 64, 72}, {48, 64, 72}, {94, 128, 144}, {94, 128, 144}, {224, 240, 255}}, //palette 1, snowflakes
  {{240, 30, 0}, {240, 30, 0}, {180, 60, 25}, {180, 60, 25}, {20, 180, 20}, {20, 20, 180}, {20, 20, 180}, {120, 160, 90}}, //palette 3 C9
  {{128,0,0}, {255,0,0}, {255,128,0}, {200,200,0}, {128,255,0}, {0,255,0}, {0,128,0}, {255,0,80}}, //palette 4, red/greens
  {{0,0,255}, {0,0,128}, {0,128,255}, {128,0,255}, {0,200,200}, {200,0,200}, {100,200,200}, {200,100,200}}, //palette 5, blues
  //{0B, 0B, 0B, 0B, 0B, 0B, 0B, 0B},
};

#define animation_instances 50
uint8_t animation_counter = 0; //which instance to put the next animation in
uint8_t animation_palette = 0; //what palette to use
uint16_t random_interval = 1200; //how much random instances trigger (of 0 - 65536, how likely is each pixel to trigger each cycle)
uint8_t animation_position[animation_instances]; //where the given animation is happening
uint8_t animation_progression[animation_instances]; //at what point in time the animation is
uint32_t animation_color[animation_instances]; //what color from the palette is used for this animation

uint8_t twinkles_rise_time = 120; //at what time rise finishes (0-255, the remainder is used to hold, the sum of the total rise/hold time is animation speed)

uint32_t animation_next_target = 15000;
uint16_t animation_next_delay = 8000;
uint16_t animation_next_counter = 0;
uint8_t animation_palette_counter = 0;

Adafruit_NeoPixel strip(led_count, led_pin, NEO_GRB + NEO_KHZ800); //uses 3*168 = 504 bytes of memory

void setup() {
  ResetAll(); //set all pixels to 0

  strip.begin();           // INITIALIZE NeoPixel strip object (REQUIRED)
  strip.show();            // Turn OFF all pixels ASAP
  strip.setBrightness(150); // Set BRIGHTNESS to about 2/5 (max = 255) to limit to max 4A

  Serial.begin(115200);

  //SetRainbowBasic();
  SetTwinklesBasic(0);
  //SetTwinklesSlow(0);
  //SetTracesBasic(0);
  //SetDropletsBasic(0);
  //SetCyclerBasic(0);
}

void loop() {
  if (millis() > animation_update_target) {
    animation_update_target = millis() + animation_update_delay;
    //update animation to next one if desired
  }
  if (millis() > pattern_update_target) {
    pattern_update_target = millis() + pattern_update_delay;
    switch (animation_state) {
      case animation_off: {
        } break;
      case animation_basic_rainbow: {
          static uint8_t rainbow_state; //initialize rainbow state
          rainbow_state += pattern_speed; //add to state
          uint8_t E_step = 256 / edge_count; //create step for each edge
          uint16_t hue_step = 65536 / (edge_size * 2); //create steps between pixel hue (*2 to keep only a section of the rainbow visible)
          uint16_t led_counter = 0;
          for (uint16_t E = 0; E < edge_count; E++) { //loop through all edges
            uint8_t temp_start = rainbow_state + (E * E_step); //create starting point for this strip
            for (uint16_t L = 0; L < edge_size; L++) {
              uint16_t temp_hue = (temp_start * 256) + (hue_step * L);
              strip.setPixelColor(led_counter, strip.gamma32(strip.ColorHSV(temp_hue)));
              led_counter++;
            }
          }
        } break;
      case animation_twinkles: {
          //randomly start new twinkles
          for (uint16_t L = 0; L < led_count; L++) { //loop through all leds
            uint16_t temp_random = random(65536);
            if (random_interval > temp_random) { //if random value is smaller than the threshold
              //Serial.println("Random hit");
              animation_position[animation_counter] = L; //set position to current led
              uint8_t temp_c_random = random(8);
              animation_color[animation_counter] = strip.Color(palette[animation_palette][temp_c_random][0], palette[animation_palette][temp_c_random][1], palette[animation_palette][temp_c_random][2]); //set color to random color from palette
              animation_progression[animation_counter] = 1; //start animation (move from 0 to 1
              animation_counter++;
              if (animation_counter >= animation_instances) { //overflow protection for led counter
                animation_counter = 0;
              }
            }
          }
        } break;
      case animation_traces: {
          //randomly start new traces
          for (uint16_t E = 0; E < edge_count; E++) { //loop through all leds
            uint16_t temp_random = random(65536);
            if (random_interval > temp_random) { //if random value is smaller than the threshold
              //Serial.println("Random hit");
              animation_position[animation_counter] = E; //set position to current led
              uint8_t temp_c_random = random(8);
              animation_color[animation_counter] = strip.Color(palette[animation_palette][temp_c_random][0], palette[animation_palette][temp_c_random][1], palette[animation_palette][temp_c_random][2]); //set color to random color from palette
              uint8_t temp_direction = random(2);
              if (temp_direction == 0) { //if movement in positive direction
                animation_progression[animation_counter] = 1; //start animation (move from 0 to 1
              }
              else { //if movement in negative direction
                animation_progression[animation_counter] = 255; //start animation (move from 0 to 255
              }
              animation_counter++;
              if (animation_counter >= animation_instances) { //overflow protection for led counter
                animation_counter = 0;
              }
            }
          }
        } break;
        case animation_droplets: {
          //randomly start new traces
          for (uint16_t E = 0; E < edge_count; E++) { //loop through all leds
            uint16_t temp_random = random(65536);
            if (random_interval > temp_random) { //if random value is smaller than the threshold
              //Serial.println("Random hit");
              animation_position[animation_counter] = E; //set position to current led
              uint8_t temp_c_random = random(8);
              animation_color[animation_counter] = strip.Color(palette[animation_palette][temp_c_random][0], palette[animation_palette][temp_c_random][1], palette[animation_palette][temp_c_random][2]); //set color to random color from palette
              animation_progression[animation_counter] = 1; //start animation (move from 0 to 1
              animation_counter++;
              if (animation_counter >= animation_instances) { //overflow protection for led counter
                animation_counter = 0;
              }
            }
          }
        } break;
        case animation_cycler: {
          static uint8_t cycler_state; //initialize rainbow state
          static uint8_t color_state;
          if (color_state >= 8){ //overflow protection for color
            color_state = 0;
          }
          strip.setPixelColor(cycler_state, strip.gamma32(strip.Color(palette[animation_palette][color_state][0], palette[animation_palette][color_state][1], palette[animation_palette][color_state][2])));
          
          cycler_state += 1; //add to state
          if (cycler_state >= led_count){ //overflow protection for led counter
            cycler_state = 0; 
            color_state ++; //if overflow, go to next color
          }
          
        } break;
    }
  }
  if (millis() > frame_update_target) {
    frame_update_target = millis() + frame_update_delay;
    switch (animation_state) {
      case animation_off: {
          AllOff();
        } break;
      case animation_basic_rainbow: {
        } break;
      case animation_twinkles: {
          for (uint8_t A = 0; A < animation_instances; A++) {
            if (animation_progression[A] > 0) { //if animation is in progress
              uint32_t temp_calc = 0;
              if (animation_progression[A] < twinkles_rise_time) { //if within rise time
                temp_calc = (256 * animation_progression[A]) / twinkles_rise_time; //calculate percentage
              }
              else { //if in hold
                temp_calc = 255; //set to full
              }
              //Serial.println(temp_calc);
              if (temp_calc > pixel_heat[animation_position[A]]) { //if new heat is higher than current heat, set higher heat to pixel (and color)
                pixel_heat[animation_position[A]] = temp_calc;
                pixel_color[animation_position[A]] = animation_color[A];
              }

              //progress animation
              if (255 - pattern_speed < animation_progression[A]) { //if animation will overflow in next time, set to 0 and stop
                animation_progression[A] = 0;
              }
              else { //progress pattern
                animation_progression[A] += pattern_speed;
              }
            } //end animation progression
          } //end for loop
        } break;
      case animation_traces: {
          for (uint8_t A = 0; A < animation_instances; A++) {
            if (animation_progression[A] > 0) { //if animation is in progress
              int32_t temp_calc;
              if (animation_progression[A] <= 128) { //if movement in the positive direction
                temp_calc = 128 / edge_size;
                temp_calc = animation_progression[A] / (temp_calc); //calculate what pixel is on
                temp_calc += (animation_position[A] * edge_size); //add the edge position to this

                //progress animation
                if (128 - pattern_speed < animation_progression[A]) { //if animation will overflow in next time, set to 0 and stop
                  animation_progression[A] = 0;
                }
                else { //progress pattern
                  animation_progression[A] += pattern_speed;
                }
              }
              else { //if movement is in negative direction
                temp_calc = animation_progression[A] - 128; //invert direction
                temp_calc = map(temp_calc, 0, 128, 0, edge_size - 1); //temp_calc / (127 / edge_size); //calculate what pixel is on
                //Serial.println(temp_calc);
                temp_calc += (animation_position[A] * edge_size); //add the edge position to this

                //progress animation
                if (128 + pattern_speed > animation_progression[A]) { //if animation will overflow in next time, set to 0 and stop
                  animation_progression[A] = 0;
                }
                else { //progress pattern
                  animation_progression[A] -= pattern_speed;
                }
              }

              //Serial.println(temp_calc);
              //set color to pixels
              if (255 > pixel_heat[animation_position[A]]) { //if new heat is higher than current heat, set higher heat to pixel (and color)
                pixel_heat[temp_calc] = 255;
                pixel_color[temp_calc] = animation_color[A];
              }
            } //end animation progression
          } //end for loop
        } break;
        case animation_droplets: {
          for (uint8_t A = 0; A < animation_instances; A++) {
            if (animation_progression[A] > 0) { //if animation is in progress
              int32_t temp_calc, temp_calc1, temp_calc2;
              temp_calc = map(animation_progression[A], 0, 256, 0, (edge_size - 1) / 2); //move progression to half of an edge
              temp_calc1 = (animation_position[A] * edge_size) + (edge_size / 2); //set calc1 to center of the edge
              temp_calc2 = temp_calc1;
              temp_calc1 += temp_calc; //higher part
              temp_calc2 -= temp_calc; //lower part

              //progress animation
              if (255 - pattern_speed < animation_progression[A]) { //if animation will overflow in next time, set to 0 and stop
                animation_progression[A] = 0;
              }
              else { //progress pattern
                animation_progression[A] += pattern_speed;
              }
            
              //Serial.println(temp_calc);
              //set color to pixels
              pixel_heat[temp_calc1] = 255;
              pixel_color[temp_calc1] = animation_color[A];
              pixel_heat[temp_calc2] = 255;
              pixel_color[temp_calc2] = animation_color[A];
              
            } //end animation progression
          } //end for loop
        } break;
        case animation_cycler: {
        } break;
    }

    //Render all pixels based on color and heat ------------------
    if (heat_animation == 1) {
      for (uint16_t L = 0; L < led_count; L++) {
        uint32_t temp_color = Convert32Bit(pixel_color[L] , pixel_heat[L]);
        //Serial.println(temp_color);
        strip.setPixelColor(L, strip.gamma32(temp_color));
      }
    }

    strip.show();

    //cool down pixels after rendering
    if (universal_cooldown == 1) { //check if all heats need to be cooled down
      for (uint16_t L = 0; L < led_count; L++) {
        if (universal_cooling > pixel_heat[L]) { //if removing cooling from heat would lead to an underflow
          pixel_heat[L] = 0;
        }
        else { //simply subtract the cooling value
          pixel_heat[L] -= universal_cooling;
        }
      }
    }
  }
  if (millis() > animation_next_target){
    animation_next_target = millis() + animation_next_delay;
    animation_palette_counter ++; //add one to palette
    if (animation_palette_counter >=  num_palettes){
      animation_palette_counter = 0;
      animation_next_counter++;
    }
    if (animation_next_counter >= 3){
      animation_next_counter = 0;
    }
    switch (animation_next_counter){
      case 0: {
        SetTwinklesBasic(animation_palette_counter);
      } break;
      case 1: {
        SetTracesBasic(animation_palette_counter);
      } break;
      case 2: {
        SetDropletsBasic(animation_palette_counter);
      } break;
    }
  }
}

void ResetAll() {
  for (uint16_t L = 0; L < led_count; L++) {
    pixel_color[L] = 0;
    pixel_heat[L] = 0;
  }
}

//takes the 8 bit value and converts it to 24 bit color. 8 bit is treated as RRRGGGBB, so blue has a narrower colorspace.
//Color is scaled linearly, but later functions do apply gamma correction
//Brightness is 0-255 and determines with a linear scale how bright to make the color
uint32_t Convert8Bit(uint8_t temp_color, uint8_t temp_brightness) {
  //Serial.print("Converting: "); Serial.print(temp_color); Serial.print(", "); Serial.println(temp_brightness);
  uint32_t temp_return = 0;
  uint8_t r = temp_color & 0B11100000;
  r = map(r, 0, 224, 0, 255);
  uint8_t g = temp_color & 0B00011100;
  g = map(g, 0, 28, 0, 255);
  uint8_t b = temp_color & 0B00000011;
  b = map(b, 0, 3, 0, 255);
  //Serial.print(r); Serial.print(", "); Serial.print(g); Serial.print(", "); Serial.println(b);

  uint32_t temp_calc = r;
  temp_calc *= temp_brightness;
  temp_calc /= 256;
  temp_return |= temp_calc;
  //Serial.print(temp_calc); Serial.print(", ");
  temp_return = temp_return << 8;

  temp_calc = g;
  temp_calc *= temp_brightness;
  temp_calc /= 256;
  temp_return |= temp_calc;
  //Serial.print(temp_calc); Serial.print(", ");
  temp_return = temp_return << 8;

  temp_calc = b;
  temp_calc *= temp_brightness;
  temp_calc /= 256;
  //Serial.println(temp_calc);
  temp_return |= temp_calc;
  return temp_return;
}

uint32_t Convert32Bit(uint32_t temp_color, uint8_t temp_brightness) { //converts a 32 bit color
  uint32_t temp_return = 0;
  uint32_t r = (temp_color >> 16) & 255;
  uint32_t g = (temp_color >> 8)  & 255;
  uint32_t b = temp_color & 255;
  //Serial.print(r); Serial.print(", "); Serial.print(g); Serial.print(", "); Serial.println(b);

  uint32_t temp_calc = r;
  temp_calc *= temp_brightness;
  temp_calc /= 256;
  temp_return |= temp_calc;
  //Serial.print(temp_calc); Serial.print(", ");
  temp_return = temp_return << 8;

  temp_calc = g;
  temp_calc *= temp_brightness;
  temp_calc /= 256;
  temp_return |= temp_calc;
  //Serial.print(temp_calc); Serial.print(", ");
  temp_return = temp_return << 8;

  temp_calc = b;
  temp_calc *= temp_brightness;
  temp_calc /= 256;
  //Serial.println(temp_calc);
  temp_return |= temp_calc;
  return temp_return;
}


void AllOff() {
  strip.clear();
}

void SetRainbowBasic() {
  strip.setBrightness(150);
  animation_state = animation_basic_rainbow;
  pattern_update_delay = 30;
  pattern_speed = 5;
  frame_update_delay = 20;
  universal_cooldown = 0;
  universal_cooling = 5;
  heat_animation = 0;
}

void SetTwinklesBasic(uint8_t temp_palette) {
  strip.setBrightness(255);
  animation_state = animation_twinkles;
  pattern_update_delay = 100;
  pattern_speed = 30;
  frame_update_delay = 40;
  animation_palette = temp_palette;
  random_interval = 2000;
  universal_cooldown = 1;
  universal_cooling = 15;
  heat_animation = 1;
}

void SetTwinklesSlow(uint8_t temp_palette) {
  strip.setBrightness(255);
  animation_state = animation_twinkles;
  pattern_update_delay = 100;
  pattern_speed = 4;
  frame_update_delay = 20;
  animation_palette = temp_palette;
  random_interval = 1000;
  universal_cooldown = 1;
  universal_cooling = 3;
  heat_animation = 1;
}

void SetTracesBasic(uint8_t temp_palette) {
  strip.setBrightness(255);
  animation_state = animation_traces;
  pattern_update_delay = 100;
  pattern_speed = 12;
  frame_update_delay = 40;
  animation_palette = temp_palette;
  random_interval = 1000;
  universal_cooldown = 1;
  universal_cooling = 10;
  heat_animation = 1;
}

void SetDropletsBasic(uint8_t temp_palette) {
  strip.setBrightness(255);
  animation_state = animation_droplets;
  pattern_update_delay = 100;
  pattern_speed = 20;
  frame_update_delay = 20;
  animation_palette = temp_palette;
  random_interval = 1200;
  universal_cooldown = 1;
  universal_cooling = 4;
  heat_animation = 1;
}

void SetCyclerBasic(uint8_t temp_palette) {
  strip.setBrightness(150);
  animation_state = animation_cycler;
  pattern_update_delay = 30;
  pattern_speed = 5;
  animation_palette = temp_palette;
  frame_update_delay = 20;
  universal_cooldown = 0;
  universal_cooling = 5;
  heat_animation = 0;
}
